/*
 *  Userspace management of snd-bt-sco
 *
 *  Copyright (c) 2003 by Jonathan Paisley <jp@dcs.gla.ac.uk>
 * 
 *  Daemon enhancements (c) 2004 by Lars Grunewaldt <lgw@dark-reality.de>
 *
 *  Based on sb16_csp/cspctl.c and hstest.c from bluez-utils/test.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <getopt.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <linux/ioctl.h>
#include <alsa/asoundlib.h>
#include <sound/btsco.h>
#include <sound/asound.h>

#include <sys/wait.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sco.h>
#include <bluetooth/rfcomm.h>


#define NOT_CONNECTED 0
#define CONNECTED 1

static volatile int terminate = 0;

static void sig_term(int sig)
{
	terminate = 1;
}


static int rfcomm_connect(bdaddr_t * src, bdaddr_t * dst, uint8_t channel)
{
	struct sockaddr_rc addr;
	int s;

	if ((s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.rc_family = AF_BLUETOOTH;
	bacpy(&addr.rc_bdaddr, src);
	addr.rc_channel = 0;
	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		close(s);
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.rc_family = AF_BLUETOOTH;
	bacpy(&addr.rc_bdaddr, dst);
	addr.rc_channel = channel;
	if (connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		close(s);
		return -1;
	}

	return s;
}

static int sco_connect(bdaddr_t * src, bdaddr_t * dst, uint16_t * handle,
		       uint16_t * mtu)
{
	struct sockaddr_sco addr;
	struct sco_conninfo conn;
	struct sco_options opts;
	int s, size;

	if ((s = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_SCO)) < 0) {
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sco_family = AF_BLUETOOTH;
	bacpy(&addr.sco_bdaddr, src);
	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		close(s);
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sco_family = AF_BLUETOOTH;
	bacpy(&addr.sco_bdaddr, dst);
	if (connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		close(s);
		return -1;
	}

	size = sizeof(conn);
	if (getsockopt(s, SOL_SCO, SCO_CONNINFO, &conn, &size) < 0) {
		close(s);
		return -1;
	}

	size = sizeof(opts);
	if (getsockopt(s, SOL_SCO, SCO_OPTIONS, &opts, &size) < 0) {
		close(s);
		return -1;
	}

	if (handle)
		*handle = conn.hci_handle;

	if (mtu)
		*mtu = opts.mtu;

	return s;
}

static void error(const char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	fprintf(stderr, "Error: ");
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	va_end(va);
}

static int bt_sco_set_fd(snd_hwdep_t * handle, int sco_fd)
{
	if (snd_hwdep_ioctl(handle,SNDRV_BT_SCO_IOCTL_SET_SCO_SOCKET,(void *) sco_fd) < 0) {
		error("unable to set fd");
		return 1;
	}
	return 0;
}

int find_hwdep_device(int *cardP,int *devP)
{
	snd_ctl_t *ctl_handle;
	snd_ctl_card_info_t *card_info;
	snd_hwdep_info_t *hwdep_info;

	int card;
	int dev;
	int err;
	char card_id[32];

	ctl_handle = NULL;

	snd_ctl_card_info_alloca(&card_info);
	snd_hwdep_info_alloca(&hwdep_info);

	for (card = 0; card < 7; card++) {
		*cardP = card;

		if (ctl_handle) {
			snd_ctl_close(ctl_handle);
			ctl_handle = NULL;
		}
		// Get control handle for selected card
		sprintf(card_id, "hw:%i", card);
		if ((err = snd_ctl_open(&ctl_handle, card_id, 0)) < 0) {
			error("control open (%s): %s", card_id,
			      snd_strerror(err));
			return -1;
		}
		// Read control hardware info from card
		if ((err = snd_ctl_card_info(ctl_handle, card_info)) < 0) {
			error("control hardware info (%s): %s", card_id,
			      snd_strerror(err));
			continue;
		}

		//if (strcmp(snd_ctl_card_info_get_driver(card_info),"BT SCO (d)"))
		//    continue;

		dev = -1;
		err = 1;
		while (1) {
			if (snd_ctl_hwdep_next_device(ctl_handle, &dev) <
			    0)
				error("hwdep next device (%s): %s",
				      card_id, snd_strerror(err));
			if (dev < 0)
				break;
			snd_hwdep_info_set_device(hwdep_info, dev);
			if (snd_ctl_hwdep_info(ctl_handle, hwdep_info) < 0) {
				if (err != -ENOENT)
					error
					    ("control hwdep info (%s): %s",
					     card_id, snd_strerror(err));
				continue;
			}
			if (snd_hwdep_info_get_iface(hwdep_info) ==
			    SNDRV_HWDEP_IFACE_BT_SCO) {
				snd_ctl_close(ctl_handle);
				*devP = dev;
				return 0;
			}
		}
	}
	if (ctl_handle)
		snd_ctl_close(ctl_handle);

	return -1;
}

static void usage(void)
{
	printf("Usage:\n" "\tbtsco <bdaddr> [channel]\n");
}

int main(int argc, char *argv[])
{
	int dev;
	int card;

	struct sigaction sa;

	fd_set rfds;
	struct timeval timeout;
	unsigned char buf[2048];
	int sel, rlen, wlen;

	bdaddr_t local;
	bdaddr_t bdaddr;
	uint8_t channel;

	char *filename;
	mode_t filemode;
	int mode = 0;
	int dd;
	int rd; // rfcomm handle
	int sd; //sco handle
	uint16_t sco_handle, sco_mtu, vs;
	char line[100];
	int volumes[2], last_volumes[2];
	
	// sco_mode is our running mode. 0 => not connect, 1 => connected
	// see NOT_CONNECTED,CONNECTED :)
	int sco_mode;

	struct pollfd pfds[10];
	int nfds;

	int i, err;

	snd_hwdep_t *handle;
	char hwdep_name[16];


	/* detect the audio device */
	if (find_hwdep_device(&card,&dev)) {
		error("Can't find device. Bail");
		return 1;
	}

	printf("Device is %d:%d\n", card, dev);

	sprintf(hwdep_name, "hw:%i,%i", card, dev);

	/* open hwdep on audio device */
	if ((err = snd_hwdep_open(&handle, hwdep_name, O_RDWR)) < 0) {
			error("btsco open (%i-%i): %s\n", card, dev,
 	      	snd_strerror(err));
 	      	return -1;
	}
	
	if (argc > 3) {
		printf("Clearing fd\n");
		bt_sco_set_fd(handle, 1);
		return 1;
	}
	
	
	/* find bdaddr */
	switch (argc) {
	case 2:
		str2ba(argv[1], &bdaddr);
		channel = 2;
		break;
	case 3:
		str2ba(argv[1], &bdaddr);
		channel = atoi(argv[2]);
		break;
	default:
		usage();
		exit(-1);
	}

	
	
	/* check voice settings. in this version we only support mu-law */
	hci_devba(0, &local);
	dd = hci_open_dev(0);
	hci_read_voice_setting(dd, &vs, 1000);
	vs = htobs(vs);
	fprintf(stderr, "Voice setting: 0x%04x\n", vs);
	close(dd);
	/*
	  MU_LAW
	if (vs != 0x0140) {
		fprintf(stderr, "The voice setting must be 0x0140\n");
		return -1;
	}
	*/

	// 16bit
	if (vs != 0x060) {
		fprintf(stderr, "The voice setting must be 0x060\n");
		return -1;
	}

	/* setup sigterm handler. we must make sure to do a clean disconnect */
	memset(&sa, 0, sizeof(sa));
	sa.sa_flags = SA_NOCLDSTOP;
	sa.sa_handler = sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT, &sa, NULL);

	sa.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);

	/* connect rfcomm control channel */
	if ((rd = rfcomm_connect(&local, &bdaddr, channel)) < 0) {
		perror("Can't connect RFCOMM channel");
		return -1;
	}

	fprintf(stderr, "RFCOMM channel connected\n");


	i = 0;

	/* set up data polling description */
	nfds = 0;
	
	/* polling data from rfcomm */
	pfds[nfds].fd = rd;
	pfds[nfds++].events = POLLIN;

    // polling data from command line - unused now
//	pfds[nfds].fd = 0;
//	pfds[nfds++].events = POLLIN;

    /* polling data from hwdep interface */
	nfds += snd_hwdep_poll_descriptors(handle, &pfds[nfds], 1);

	last_volumes[0] = last_volumes[1] = 0;

	/* we are not yet connected */
	sco_mode=NOT_CONNECTED;
	while (!terminate) {

		if (poll(pfds, nfds, 1000) > 0) {
			short revents;

			/* Volume polling (sound card) */
			if (!snd_hwdep_poll_descriptors_revents
			    (handle, &pfds[nfds - 1], 1, &revents)
			    && revents & POLLIN) {
				int len;

				len =
				    snd_hwdep_read(handle, volumes,
						   sizeof(volumes));
				if (len == sizeof(volumes)) {
					printf
					    ("speaker volume: %d mic volume: %d\n",
					     volumes[0], volumes[1]);
					if (volumes[0] != last_volumes[0]) {
						sprintf(line, "+VGS=%d\r",
							volumes[0]);
						write(rd, line,
						      strlen(line));
					}
					if (volumes[1] != last_volumes[1]) {
						sprintf(line, "+VGM=%d\r",
							volumes[1]);
						write(rd, line,
						      strlen(line));
					}
					memcpy(last_volumes, volumes,
					       sizeof(volumes));
				}
			}

			// control transmission events for volume and channel control
			if (pfds[0].revents & POLLIN) {
				memset(buf, 0, sizeof(buf));
				rlen = read(rd, buf, sizeof(buf) - 1);
				if (rlen > 0) {
					fprintf(stderr, "recieved %s\n", buf);
					/* tell them we recieved */
					wlen = write(rd, "\r\nOK\r\n", 6);
					
					if (strstr(buf, "AT+BVRA=") || strstr(buf,"AT+CKPD=200")) {
						/* mini state machine: handle connect/disconnect */
						switch (sco_mode) {
							case NOT_CONNECTED:
								fprintf(stderr, "opened hwdep\n");
								/* connect sco stream */
								if ((sd = sco_connect(&local, &bdaddr, &sco_handle, &sco_mtu)) < 0) {
									
									perror("Can't connect SCO audio channel\n");
								} else {
									fprintf(stderr, "connected SCO channel\n");
									//	write(rd, "RING\r\n", 6);
									printf("Setting sco fd\n");
									bt_sco_set_fd(handle, sd);
									
									printf("Done setting sco fd\n");
									sco_mode=CONNECTED;
								}
								
								
								break;
							case CONNECTED:
							        /* close bt_sco audio handle */
								bt_sco_set_fd(handle, -1);
								/* disconnect SCO stream */
								close(sd);
								fprintf(stderr, "disconnected SCO channel\n");
								
								sco_mode=NOT_CONNECTED;
								break;
						}
					}
					
					
					
					if (sscanf
					    (buf, "AT+VGS=%d",
					     &volumes[0]) == 1) {
						fprintf(stderr,
							"Sending up speaker change %d\n",
							volumes[0]);
						snd_hwdep_write(handle,
								volumes,
								sizeof
								(volumes));
								
					}
					if (sscanf
					    (buf, "AT+VGM=%d",
					     &volumes[1]) == 1) {
						fprintf(stderr,
							"Sending up microphone change %d\n",
							volumes[1]);
						snd_hwdep_write(handle,
								volumes,
								sizeof
								(volumes));
								
					}
				}
			}

			/* Just for testing; handled by kernel driver */
			
		if (0 && FD_ISSET(sd, &rfds)) {
				memset(buf, 0, sizeof(buf));
				rlen = read(sd, buf, sizeof(buf));
				write(sd, buf, rlen);

				i++;

				if (i % 15 == 0) {
					printf("rlen: %d\n", rlen);
				}
			}

		}

	}

	if (sco_mode==CONNECTED) {
		close(sd);

		bt_sco_set_fd(handle, -1);
		
	}

	sleep(1);
	close(rd);

	snd_hwdep_close(handle);

	return 0;
}
